-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
local mathutil = require("mathutil")
local debugutil = require("debugutil")("ecvt", true)
local constants = {
  rpmToAV = 0.104719755,
  avToRPM = 9.549296596425384,
  torqueToHpCoef = 7120.54016,
  torquePowerCoef = 9.5488
}

M.outputPorts = {[1] = true}
-- "viscouscoupling" below is a hack-around to allow a second gearbox to be piggybacked onto this one for insane hax
M.deviceCategories = {gearbox = true, viscouscoupling = true}
M.requiredExternalInertiaOutputs = {1}

local max = math.max
local min = math.min
local abs = math.abs
local floor = math.floor
local clamp = mathutil.clamp
local sign = mathutil.sign
local maprange = mathutil.maprange

-- ecvtGearbox Prototype
local ecvt = {}
ecvt.__index = ecvt

function ecvt:velocityUpdate(dt)
  -- ring gear is connected directly to output reduction gears
  self.ringGearAV = self.outputAV1 * self.finalGearRatio
  
  -- update planet carrier AV based on existing input AV and sun gear torque
  local adjustedParentAV = self.parent[self.parentOutputAVName] * self.inputGearRatio
  local adjustedParentTorque = self.parent[self.parentOutputTorqueName] / self.inputGearRatio - self.friction * min(max(adjustedParentAV, -1), 1)
  local carrierTorqueFromSun = self.motorGenerator1Torque * ((self.ringGearTeeth + self.sunGearTeeth) / self.sunGearTeeth)
  local carrierTorque = adjustedParentTorque + carrierTorqueFromSun
  
  self.planetCarrierAV = self.planetCarrierAV + carrierTorque * self.invCumulativeInertia * dt
  self.inputAV = self.planetCarrierAV / self.inputGearRatio
  self.parent[self.parentOutputAVName] = self.inputAV
  
  -- figure out sun gear from other two
  self.sunGearAV = ((self.sunGearTeeth + self.ringGearTeeth) * self.planetCarrierAV - (self.ringGearTeeth * self.ringGearAV)) / self.sunGearTeeth
end

function ecvt:torqueUpdate(dt)
  local reverseGearRatioCoef = self.reverseGearRatioCoef
  local oneWayTorque = self.oneWayTorqueSmoother:get(
    min(max(self.oneWayViscousCoef * self.outputAV1, -self.oneWayViscousTorque), self.oneWayViscousTorque)
  )

  self.oneWayTorqueSmoother:set(self.outputAV1 * reverseGearRatioCoef < 0 and oneWayTorque or 0)
  oneWayTorque = self.oneWayTorqueSmoother:value() * reverseGearRatioCoef
  
  -- input torque from engine
  local rawCarrierAV = self.parent[self.parentOutputAVName] * self.inputGearRatio
  local rawCarrierTorque = self.parent[self.parentOutputTorqueName] / self.inputGearRatio - self.friction * min(max(rawCarrierAV, -1), 1)
  
  self.adjustedInputTorque = rawCarrierTorque
  
  -- motor torques
  local mg1AV = abs(self.sunGearAV)
  local mg2AV = abs(self.ringGearAV)
  local mg1MaxTorque = self:getMaxMG1Torque(self.motorGenerator1TorqueDemand < 0)
  local mg2MaxTorque = self:getMaxMG2Torque(self.motorGenerator2TorqueDemand < 0)
  local mg1Torque = 0
  local mg2Torque = 0
  
  if self.mode ~= "neutral" then
    if self.motorGenerator1TorqueDemand < 0 then
      mg1Torque = -sign(self.sunGearAV) * min(abs(self.motorGenerator1TorqueDemand), self:getMaxMG1RegenTorque())
    else
      mg1Torque = self.motorGenerator1TorqueDemand * self.motorGenerator1TorqueDirection
    end
    
    if self.motorGenerator2TorqueDemand < 0 then -- regenerative braking
      mg2Torque = -sign(self.ringGearAV) * min(abs(self.motorGenerator2TorqueDemand), self:getMaxMG2RegenTorque())
    else
      mg2Torque = self.motorGenerator2TorqueDemand * reverseGearRatioCoef
    end
  end
  
  -- apply limits
  mg1Torque = clamp(mg1Torque, -mg1MaxTorque, mg1MaxTorque)
  mg2Torque = clamp(mg2Torque, -mg2MaxTorque, mg2MaxTorque)
  
  self.motorGenerator1Torque = self.mg1TorqueSmoother:get(mg1Torque, dt)
  self.motorGenerator2Torque = self.mg2TorqueSmoother:get(mg2Torque, dt)
  
  -- motor powers
  local mg1PowerCoef = self.motorGenerator1TorqueDemand < 0 and 1 or self:getPowerCoef(self.sunGearAV)
  local mg2PowerCoef = self.motorGenerator2TorqueDemand < 0 and 1 or self:getPowerCoef(self.ringGearAV)
  
  self.motorGenerator1Power = mg1PowerCoef * abs(mathutil.motorTorqueToKw(mg1Torque, mg1AV)) * sign(self.motorGenerator1TorqueDemand)
  self.motorGenerator2Power = mg2PowerCoef * abs(mathutil.motorTorqueToKw(mg2Torque, mg2AV)) * sign(self.motorGenerator2TorqueDemand)
  
  if abs(self.outputAV1) < 100 then
    self.parkEngaged = 1
  end
  
  -- output torque is the same as ring gear torque, which is a sum of
  -- the torque of MG2 plus the ratio-adjusted torques from the input
  -- and MG1
  local ringTorqueFromCarrier = self:getRingTorqueFromCarrierTorque(rawCarrierTorque)
  local ringTorque = self.motorGenerator2Torque - ringTorqueFromCarrier
  
  self.outputTorque1 = ringTorque * self.finalGearRatio
  
  if self.mode == "park" then
    self.parkClutchAngle = min(max(self.parkClutchAngle + self.outputAV1 * dt, -self.maxParkClutchAngle), self.maxParkClutchAngle)
    self.outputTorque1 = self.outputTorque1 + -self.parkClutchAngle * self.parkLockSpring * self.parkEngaged * self.lockCoef
  end
  
  self.inputEngaged = self.mode ~= "neutral" and not self.inputForceDisengaged
  
  if not _G.gbDebugOff then
    local function formatAv(av)
      local dispRpm = floor(av * constants.avToRPM * 100) / 100
      local dispAv = floor(av * 100) / 100
      
      return dispRpm .. " (" .. dispAv .. ")"
    end
    
    debugutil.putValue("MG1 Torque", self.motorGenerator1Torque)
    debugutil.putValue("MG1 Max Torque", mg1MaxTorque)
    debugutil.putValue("MG1 Torque Demand", self.motorGenerator1TorqueDemand)
    debugutil.putValue("MG1 Torque Coef", self.motorGenerator1TorqueCoef)
    debugutil.putValue("MG1 Power", self.motorGenerator1Power)
    debugutil.putValue("MG2 Torque", self.motorGenerator2Torque)
    debugutil.putValue("MG2 Max Torque", mg2MaxTorque)
    debugutil.putValue("MG2 Torque Demand", self.motorGenerator2TorqueDemand)
    debugutil.putValue("MG2 Torque Coef", self.motorGenerator2TorqueCoef)
    debugutil.putValue("MG2 Power", self.motorGenerator2Power)
    debugutil.putValue("Sun Gear RPM (AV)", formatAv(self.sunGearAV))
    debugutil.putValue("Carrier RPM (AV)", formatAv(self.planetCarrierAV))
    debugutil.putValue("Ring Gear RPM (AV)", formatAv(self.ringGearAV))
    debugutil.putValue("Output Torque", self.outputTorque1)
  else
    debugutil.putValue("MG1 Torque", nil)
    debugutil.putValue("MG1 Max Torque", nil)
    debugutil.putValue("MG1 Torque Demand", nil)
    debugutil.putValue("MG1 Torque Coef", nil)
    debugutil.putValue("MG1 Power", nil)
    debugutil.putValue("MG2 Torque", nil)
    debugutil.putValue("MG2 Max Torque", nil)
    debugutil.putValue("MG2 Torque Demand", nil)
    debugutil.putValue("MG2 Torque Coef", nil)
    debugutil.putValue("MG2 Power", nil)
    debugutil.putValue("Sun Gear RPM (AV)", nil)
    debugutil.putValue("Carrier RPM (AV)", nil)
    debugutil.putValue("Ring Gear RPM (AV)", nil)
    debugutil.putValue("Output Torque", nil)
  end
end

function ecvt:validate()
  if not self.parent.deviceCategories.clutchlike then
    log("E", "ecvtGearbox.validate", "Parent device is not a clutch-like device...")
    log("E", "ecvtGearbox.validate", "Actual parent:")
    log("E", "ecvtGearbox.validate", powertrain.dumpsDeviceData(self.parent))
    return false
  end
  
  if not self.parent.parent or not self.parent.parent.deviceCategories.engine then
    log("E", "ecvtGearbox.validate", "Clutch parent's parent device is not an engine...")
    log("E", "ecvtGearbox.validate", "Actual clutch parent's parent:")
    log("E", "ecvtGearbox.validate", powertrain.dumpsDeviceData(self.parent.parent))
    return false
  end
  
  return true
end

function ecvt:setMode(mode)
  self.mode = mode
  self.reverseGearRatioCoef = mode == "reverse" and -1 or 1
end

function ecvt:setGearRatio(ratio)
end

function ecvt:setInputDisengaged(inputDisengaged)
  self.inputForceDisengaged = inputDisengaged
end

function ecvt:setRequestedMG1Torque(requestedTorque)
  self.motorGenerator1TorqueDemand = requestedTorque
end

function ecvt:setRequestedMG2Torque(requestedTorque)
  self.motorGenerator2TorqueDemand = requestedTorque
end

function ecvt:setMG1PowerOverride(maxPowerOverride)
  self.maxMG1PowerOverride = maxPowerOverride
end

function ecvt:setMG2PowerOverride(maxPowerOverride)
  self.maxMG2PowerOverride = maxPowerOverride
end

function ecvt:setMG1TorqueDirection(direction)
  self.motorGenerator1TorqueDirection = (direction == -1) and -1 or 1
end

function ecvt:getRingTorqueFromSunTorque(sunTorque)
  return sunTorque * (self.ringGearTeeth / self.sunGearTeeth)
end

function ecvt:getRingTorqueFromCarrierTorque(carrierTorque)
  return -carrierTorque * self.ringGearTeeth / (self.ringGearTeeth + self.sunGearTeeth)
end

function ecvt:getCarrierTorqueFromSunTorque(sunTorque)
  return -sunTorque * (self.ringGearTeeth + self.sunGearTeeth) / self.sunGearTeeth
end

function ecvt:getCarrierTorqueFromRingTorque(ringTorque)
  return -ringTorque * (self.ringGearTeeth + self.sunGearTeeth) / self.ringGearTeeth
end

function ecvt:getSunTorqueFromRingTorque(ringTorque)
  return ringTorque * (self.sunGearTeeth / self.ringGearTeeth)
end

function ecvt:getSunTorqueFromCarrierTorque(carrierTorque)
  return -carrierTorque * self.sunGearTeeth / (self.ringGearTeeth + self.sunGearTeeth)
end

function ecvt:setLock(enabled)
  self.lockCoef = enabled and 0 or 1
end

function ecvt:calculateInertia()
  local outputInertia = 0
  
  if self.children and #self.children > 0 then
    local child = self.children[1]
    
    outputInertia = child.cumulativeInertia
  end
  
  -- gear ratio of output to sun gear (which acts as virtual mass)
  local sunCarrierRatio = (self.ringGearTeeth + self.sunGearTeeth) / self.sunGearTeeth
  local sunCarrierInertia = self.sunGearInertia / (sunCarrierRatio * sunCarrierRatio)
  local totalInertia = self.carrierInertia + sunCarrierInertia

  self.cumulativeInertia = totalInertia / (self.inputGearRatio * self.inputGearRatio)
  self.invCumulativeInertia = 1 / self.cumulativeInertia

  self.parkLockSpring = self.jbeamData.parkLockSpring or (powertrain.stabilityCoef * powertrain.stabilityCoef * outputInertia * 0.5) --Nm/rad
  self.maxParkClutchAngle = self.parkLockTorque / self.parkLockSpring

  self.cumulativeGearRatio = 1
  self.maxCumulativeGearRatio = 1
end

function ecvt:getMaxMG1Torque(overrideOrIgnore)
  local mg1AV = max(abs(self.sunGearAV), 1e-6)
  local maxPower = self.maxMG1Power
  
  if type(overrideOrIgnore) == "number" then
    maxPower = overrideOrIgnore
  elseif overrideOrIgnore == false and type(self.maxMG1PowerOverride) == "number" then
    maxPower = self.maxMG1PowerOverride
  end
  
  local maxTorque = mathutil.motorKwToTorque(maxPower, mg1AV)
  
  return min(maxTorque, self.maxMG1Torque)
end

function ecvt:getMaxMG1RegenTorque()
  local maxTorque = self:getMaxMG1Torque(true)
  
  return maprange(abs(self.sunGearAV), 0, self.generatorMinFullLoadAV, 0, maxTorque)
end

function ecvt:getMaxMG2Torque(overrideOrIgnore)
  local mg2AV = max(abs(self.ringGearAV), 1e-6)
  local maxPower = self.maxMG2Power
  
  if type(overrideOrIgnore) == "number" then
    maxPower = overrideOrIgnore
  elseif overrideOrIgnore == false and type(self.maxMG2PowerOverride) == "number" then
    maxPower = self.maxMG2PowerOverride
  end
  
  local maxTorque = mathutil.motorKwToTorque(maxPower, mg2AV)
  
  return min(maxTorque, self.maxMG1Torque)
end

function ecvt:getMaxMG2RegenTorque()
  local maxTorque = self:getMaxMG2Torque(true)
  
  return maprange(abs(self.ringGearAV), self.regenFadeMinAV, self.regenFadeMaxAV, 0, maxTorque)
end

function ecvt:getPowerCoef(motorAv)
  return maprange(abs(motorAv), 0, self.minPowerAV, self.startingPowerCoef, 1)
end

function ecvt:getCarrierAVFromSunAV(sunAV)
  return (self.sunGearTeeth * sunAV + self.ringGearTeeth * self.ringGearAV) / (self.sunGearTeeth + self.ringGearTeeth)
end

function ecvt:getMinCarrierAV()
  return self:getCarrierAVFromSunAV(self.minMG1AV)
end

function ecvt:getMaxCarrierAV()
  return self:getCarrierAVFromSunAV(self.maxMG1AV)
end

function ecvt:updateGFX(dt)
  if streams.willSend("ecvt") then
    gui.send("ecvt", {
      ringAv = self.ringGearAV,
      sunAv = self.sunGearAV,
      carrierAv = self.planetCarrierAV,
      ringTeeth = self.ringGearTeeth,
      sunTeeth = self.sunGearTeeth,
      planetsTeeth = self.planetGearTeeth
    })
  end
end

function ecvt:reset()
  local jbeamData = self.jbeamData
  self.gearRatio = jbeamData.gearRatio
  self.friction = jbeamData.friction or 0
  self.cumulativeInertia = 1
  self.cumulativeGearRatio = 1
  self.maxCumulativeGearRatio = 1

  self.outputAV1 = 0
  self.inputAV = 0
  self.outputTorque1 = 0
  self.planetCarrierAV = 0
  self.isBroken = false
  self.ringGearAV = 0

  self.lockCoef = 1
  self.reverseGearRatioCoef = 1
  self.parkClutchAngle = 0
  self.inputEngaged = true

  --one way viscous coupling (prevents rolling backwards)
  self.oneWayTorqueSmoother:reset()
end

local function new(jbeamData)
  local device = {
    deviceCategories = shallowcopy(M.deviceCategories),
    requiredExternalInertiaOutputs = shallowcopy(M.requiredExternalInertiaOutputs),
    outputPorts = shallowcopy(M.outputPorts),
    name = jbeamData.name,
    type = jbeamData.type,
    inputName = jbeamData.inputName,
    inputIndex = jbeamData.inputIndex,
    gearRatio = jbeamData.gearRatio,
    finalGearRatio = jbeamData.finalGearRatio or 1.9,
    inputGearRatio = jbeamData.inputGearRatio or 1.0,
    friction = jbeamData.friction or 200,
    cumulativeInertia = 1,
    cumulativeGearRatio = 1,
    maxCumulativeGearRatio = 1,
    isPhysicallyDisconnected = true,
    
    outputAV1 = 0,
    inputAV = 0,
    sunGearAV = 0,
    planetCarrierAV = 0,
    ringGearAV = 0,
    
    outputTorque1 = 0,
    adjustedInputTorque = 0,
    motorGenerator1TorqueDemand = 0,
    motorGenerator1Torque = 0,
    motorGenerator1TorqueDirection = 1,
    motorGenerator2TorqueDemand = 0,
    motorGenerator2Torque = 0,
    
    motorGenerator1Power = 0,
    motorGenerator2Power = 0,
    
    isBroken = false,
    lockCoef = 1,
    reverseGearRatioCoef = 1,
    gearCount = 1,
    minGearIndex = 1,
    maxGearIndex = 1,
    gearIndex = 1,
    sunGearTeeth = jbeamData.sunGearTeeth or 40,
    planetGearTeeth = jbeamData.planetGearTeeth or 23,
    ringGearTeeth = jbeamData.ringGearTeeth or 86,
    sunGearInertia = jbeamData.sunGearInertia or 0.0483,
    carrierInertia = jbeamData.carrierInertia or 0.1202,
  }
  
  setmetatable(device, ecvt)

  --gearbox park locking clutch
  device.parkClutchAngle = 0
  device.parkLockTorque = jbeamData.parkLockTorque or 1000 --Nm

  --one way viscous coupling (prevents rolling backwards)
  device.oneWayViscousCoef = jbeamData.oneWayViscousCoef or 0
  device.oneWayViscousTorque = jbeamData.oneWayViscousTorque or device.oneWayViscousCoef * 25
  device.oneWayTorqueSmoother = newExponentialSmoothing(jbeamData.oneWayViscousSmoothing or 50)

  --electric stuff
  device.inputEngaged = true -- false to disengage engine (as if in neutral)
  device.inputForceDisengaged = false
  device.mg1ZeroSpeedMaxTorqueCoef = jbeamData.mg1ZeroSpeedMaxTorqueCoef or 3.0
  device.maxMG1AV = jbeamData.maxMG1AV or 680 -- AV (speed at which torque fades to 0)
  device.minMG1AV = jbeamData.minMG1AV or -device.maxMG1AV
  device.maxMG1Power = jbeamData.maxMG1Power or 75
  device.maxMG1PowerOverride = nil
  device.maxMG1Torque = mathutil.motorKwToTorque(device.maxMG1Power, device.maxMG1AV / 2) * device.mg1ZeroSpeedMaxTorqueCoef
  device.mg2ZeroSpeedMaxTorqueCoef = jbeamData.mg2ZeroSpeedMaxTorqueCoef or 3.0
  device.maxMG2AV = jbeamData.maxMG2AV or 850 -- AV (speed at which torque fades to 0)
  device.minMG2AV = jbeamData.minMG2AV or -device.maxMG2AV
  device.maxMG2Power = jbeamData.maxMG2Power or 85
  device.maxMG2PowerOverride = nil
  device.maxMG2Torque = mathutil.motorKwToTorque(device.maxMG2Power, device.maxMG2AV / 2) * device.mg2ZeroSpeedMaxTorqueCoef
  device.minPowerAV = jbeamData.minPowerAV or 10
  device.startingPowerCoef = jbeamData.startingPowerCoef or 2.5
  device.regenFadeMinAV = jbeamData.regenFadeMinAV or 50 -- AV
  device.regenFadeMaxAV = jbeamData.regenFadeMaxAV or 350 -- AV
  device.generatorMinFullLoadAV = jbeamData.generatorMinFullLoadAV or 30 -- AV
  device.smoothEngageFactor = jbeamData.smoothEngageFactor or 2.5
  device.mg1TorqueSmoother = newTemporalSmoothingNonLinear(jbeamData.mg1TorqueSmoothingIn or 16.65, jbeamData.mg1TorqueSmoothingOut or 16.85)
  device.mg2TorqueSmoother = newTemporalSmoothingNonLinear(jbeamData.mg2TorqueSmoothingIn or 5.65, jbeamData.mg2TorqueSmoothingOut or 5.85)

  if jbeamData.gearboxNode_nodes and type(jbeamData.gearboxNode_nodes) == "table" then
    device.transmissionNodeID = jbeamData.gearboxNode_nodes[1]
  end

  device.breakTriggerBeam = jbeamData.breakTriggerBeam
  if device.breakTriggerBeam and device.breakTriggerBeam == "" then
    --get rid of the break beam if it's just an empty string (cancellation)
    device.breakTriggerBeam = nil
  end

  device.jbeamData = jbeamData

  return device
end

M.new = new

return M
